﻿/* impro16C.ahk -- import programs to virtual HP16C */
/* Use: a) import a program from listing specified  */
/*         as argument of impro16C.ahk              */
/*      b) add few keys on top of virtual 16C       */
/*         w (west) = f-shift, e (east) = g-shift,  */
/*         middle mouse button, Ins, and F12 =      */
/*         = select a file to import,               */
/*         Ctrl+c = copy display to clipboard       */
/*         Ctrl+v = paste value from clipboard      */
/*      c) quit when virual 16C is shut down        */
/* QPL 1.0 applies, see end of file. (c) 2023 by M. */
/* "Hotel-Papa-Four-One-Charly-Xray-&-web-dot-de"   */
/* Initial version for virtual 15C      M. 10/20/23 */
/* Ctrl+c added, alas it's too slow     M. 10/29/23 */
/* Modified to use with virtual 16C     M. 12/16/23 */
/* Paste in PRGM mode assumes hex mode  M. 12/16/23 */
#Requires AutoHotkey v2.0
#SingleInstance Force     ; no 2nd instance, use youngest script
SetTitleMatchMode 2       ; 1 and 3 fail, but why?

/* if no virtual 16C running try to start it */
if WinExist("V16c ahk_class #32770")
   WinActivate
else
{
   Run "C:\PRGM\HP-Emulators\V16c\V16c.exe"
   if NOT WinWaitActive("V16c ahk_class #32770", , 2)
   {
      dc := MsgBox("Virtual 16C not up and running within 2 seconds."
             "`nAHK script ends now.", "Useless, totally", "O Iconx")
      ExitApp 9
   }
}
sesova()                ; set some vars

/* if script got an argument, assume it's as PRGM listing */
if A_Args.Length = 1
   impro(A_Args[1])

/* have a watchdog for HP 16C virtual Calculator */
SetTimer quitwhendue, 360

#HotIf WinActive("V16c ahk_class #32770")
e::RAlt
w::
{
   if umedski()         ;  if medium skin in use
      Click 149, 433
}
MButton::
Ins::
F12::
{
   fid := FileSelect(3,
    , "Pick a HP16C program listing to import"
    , "16C PRGM (*.txt; *.16C; *.16Clst)")
   if fid               ; if not aborted
      impro(fid)
}
*LCtrl::return
*RCtrl::return
#HotIf HotIfCond.CtrlV
*c::copydisplay()
*v::pastevalue()

#HotIf

Class HotIfCond {
   static CtrlV {
      get {
         return getKeyState("Ctrl","P")
          && WinActive("V16c ahk_class #32770")
      }
   }
}

/* * *  subroutines  * * */
quitwhendue()
{
   if not WinExist("V16c ahk_class #32770")
      ExitApp 9
}

/* analyse display content and put result to clipboard */
copydisplay()
{
   if umedski()         ; only go if medium skin in use
   {
      if PixelGetColor(157, 69) = 0x000000
         disco := '-'
      else disco := ' '
      Loop 10  ; ten digits only, the sign of the mantissa is an annunciator
      {
         offset := A_Index * 34 + 100
         if PixelGetColor(offset + 40, 76) = 0x000000
         {
            if PixelGetColor(offset + 55, 76) = 0x000000
            {
               if PixelGetColor(offset + 55, 59) = 0x000000
               {
                  if PixelGetColor(offset + 47, 69) = 0x000000
                  {
                     if PixelGetColor(offset + 47, 84) = 0x000000
                     {
                        if PixelGetColor(offset + 48, 52) = 0x000000
                           disco := disco '8'
                        else disco := disco 'd'
                     }
                     else disco := disco 'A'
                  }
                  else disco := disco '0'
               }
               else
               {
                  if PixelGetColor(offset + 52, 52) = 0x000000
                     disco := disco '6'
                  else
                  {
                     if PixelGetColor(offset + 40, 60) = 0x000000
                     {
                        if PixelGetColor(offset + 47, 84) = 0x000000
                           disco := disco 'b'
                        else disco := disco 'h'
                     }
                     else disco := disco 'o'
                  }
               }
            }
            else
            {
               if PixelGetColor(offset + 45, 84) = 0x000000
               {
                  if PixelGetColor(offset + 55, 61) = 0x000000
                     disco := disco '2'
                  else
                  {
                     if PixelGetColor(offset + 48, 69) = 0x000000
                        disco := disco 'E'
                     else disco := disco 'C'
                  }
               }
               else
               {
                  if PixelGetColor(offset + 55, 61) = 0x000000
                     disco := disco 'P'
                  else
                  {
                     if PixelGetColor(offset + 41, 60) = 0x000000
                        disco := disco 'F'
                     else disco := disco 'r'
                  }
               }
            }
         }
         else
         {
            if PixelGetColor(offset + 48, 53) = 0x000000
            {
               if PixelGetColor(offset + 42, 61) = 0x000000
               {
                  if PixelGetColor(offset + 55, 61) = 0x000000
                     disco := disco '9'
                  else disco := disco '5'
               }
               else
               {
                  if PixelGetColor(offset + 46, 83) = 0x000000
                     disco := disco '3'
                  else disco := disco '7'
               }
            }
            else
            {
               if PixelGetColor(offset + 55, 78) = 0x000000
               {
                  if PixelGetColor(offset + 47, 69) = 0x000000
                     disco := disco '4'
                  else disco := disco '1'
               }
               else
               {
                  if PixelGetColor(offset + 47, 69) = 0x000000
                     disco := disco '-'
                  else disco := disco ' '
               }
            }
         }
         if PixelGetColor(offset + 64, 84) = 0x000000
         {
            if PixelGetColor(offset + 62, 89) = 0x000000
               disco := disco ","
            else disco := disco "."
         }
      }
      A_Clipboard := disco
      SoundBeep(371, 99)   ; indicate "done"
   }   
}

/* paste value from clipboard */
pastevalue()
{
   if umedski()         ; only go if medium skin in use
   {
      Loop Parse A_Clipboard, "`n", "`r"
      {
         if StrLen(A_LoopField)
         {
            xm := 0
            pm := (PixelGetColor(477, 101) = 0x000000)
            if hexmode() OR pm
            {
               xm := 1
               rx := "[[:xdigit:]]+"
            }
            else rx := "i)-?(\d[\., ']?)+(e-?\d{1,2})?"
            if RegExMatch(A_LoopField, rx, &r)
            {
               r := StrReplace(r[], "'")
               if InStr(r, ".", 1, , -1) < InStr(r, ",", 1, , -1)
               {
                  r := StrReplace(r, ".")
                  r := StrReplace(r, ",", ".")
               }
               if Ord(r) = 45
                  r := SubStr(r, 2, 1) "-" SubStr(r, 3)
               if not pm
                  Send "{ESC 2}"
               Loop Parse, r, , ", "
               {
                  if xm
                     pushkey(A_LoopField)
                  else
                  {
                     Switch A_LoopField, 0
                     {
                        case "-":
                           pushkey("49")
                        case "E":
                           pushkey("42 49")
                        Default:
                           Send A_LoopField
                     }
                  }
               }
               if not pm
                  Send "{ESC 2}"
            }
            else SoundBeep
         }
         else SoundBeep
         break          ; use 1st record only
      }
   }
}

/* in hex mode? */
hexmode()
{
   if not PixelGetColor(480, 77) = 0x000000
      Return 0          ; last digit either blank, 1, 3, 4, 5, 7, or 9
   if PixelGetColor(486, 84) = 0x000000
      Return 0          ; last digit either 0, 2, 6, 8, d, o, or b
   Return 1             ; highly likely in hex mode
}

/* medium skin in use? */
umedski()
{
   WinGetPos , , &W, , "V16c ahk_class #32770"
   if (W - 806)
   {
      MsgBox "Script prepared to work with medium skin only.`nExecuting ends."
      Return 0
   }
   Return 1
}

/* import program */
impro(fid)
{
   listing := FileRead(fid)
   if listing = ""
   {
      MsgBox "Can't read file " fid
      Return            ; quit
   }

/* Virtual 16C is up -- with which skin? */
   if not umedski()
      Return

/* Virtual 16C is on or off, if off switch it on */
   Loop
   {
      if PixelSearch(&Px, &Py, 175, 49, 480, 84, 0x000000)
         Break
      if A_Index > 2
         {
            MsgBox "Unable to switch on HP 16C Virtual Calculator.`nExecuting ends."
            Return
         }
      pushkey(41)
      Sleep 80          ; wait patiently
   }

/* Virtual 16C is not in PRGM mode? */
   if not (PixelGetColor(477, 101) = 0x000000)
      pushkey("43 31")

/* process input file line by line */
   Loop parse, listing, "`n", "`r"        ; thinly sliced
   {
      line := A_LoopField
      if InStr(line, "#!", 0, 1, 1) = 1
      {
         if InStr(line, "16C")
            Continue
         MsgBox "According to ahsh-bang not a 16C program:`n" line
         Return
      }
      if InStr(line, "*", 0, 1, 1) = 1    ; comments
         Continue
      if Trim(line) = ""                  ; .bl
         Continue
      line := SubStr(line, 5, 8)
      if line = "+....1.."                ; scale
         Continue
      line := StrReplace(line, ".", "48,")
      line := StrReplace(line, ",", A_Space)
      pushkey(Trim(line))                  ; missing REXX space()
   }
}

/* Push Key or list of keys */
pushkey(lok)
{
/* no case filter required in AHK, all specials set in kmx. */
   Loop Parse, lok, A_Space
   {
      if StrLen(Trim(A_LoopField))     ; bare IF fails with string "0"
         Send kmx[Trim(A_LoopField)]
   }
}

/* set some vars */
sesova()
{
   global
   kmx := Map(
     "A",'{F1}'     ,"21",'{F7}'      ,"31",'{Click 76, 353}',"41",'{Esc}'
    ,"b",'{F2}'     ,"22",'{F8}'      ,"32",'g'         ,"42",'{Click 149, 433}'
    ,"C",'{F3}'     ,"23",'{F9}'      ,"33",'{Down}'    ,"43",'{RAlt}'
    ,"d",'{F4}'     ,"24",'{F10}'     ,"34",'{Left}'    ,"44",'s'
    ,"E",'{F5}'     ,"25",'{F11}'     ,"35",'{BS}'      ,"45",'r'
    ,"F",'{F6}'     ,"26",'u'         ,"36",'{Enter}'
    ,"7",'{Numpad7}',"4" ,'{Numpad4}' ,"1" ,'{Numpad1}' ,"0" ,'{Numpad0}'
    ,"8",'{Numpad8}',"5" ,'{Numpad5}' ,"2" ,'{Numpad2}' ,"48",'{NumpadDot}'
    ,"9",'{Numpad9}',"6" ,'{Numpad6}' ,"3" ,'{Numpad3}' ,"49",'{+}'
    ,"10",'{NumpadDiv}',"20",'{NumpadMult}',"30",'{NumpadSub}',"40",'{NumpadAdd}')
   kmx.Default := 'unknown'
}

/* * * * *  THE Q PUBLIC LICENSE version 1.0  * * * * *
Copyright (C) 1999 Trolltech AS, Norway.

Everyone is permitted to copy and distribute this license document. The intent of this license is to establish freedom to share and change the software regulated by this license under the open source model.

This license applies to any software containing a notice placed by the copyright holder saying that it may be distributed under the terms of the Q Public License version 1.0. Such software is herein referred to as the Software. This license covers modification and distribution of the Software, use of third-party application programs based on the Software, and development of free software which uses the Software.

Granted Rights

  1.
    You are granted the non-exclusive rights set forth in this license provided you agree to and comply with any and all conditions in this license. Whole or partial distribution of the Software, or software items that link with the Software, in any form signifies acceptance of this license.
  2.
    You may copy and distribute the Software in unmodified form provided that the entire package, including - but not restricted to - copyright, trademark notices and disclaimers, as released by the initial developer of the Software, is distributed.
  3.
    You may make modifications to the Software and distribute your modifications, in a form that is separate from the Software, such as patches. The following restrictions apply to modifications:
    a. Modifications must not alter or remove any copyright notices in the Software.
    b. When modifications to the Software are released under this license, a non-exclusive royalty-free right is granted to the initial developer of the Software to distribute your modification in future versions of the Software provided such versions remain available under these terms in addition to any other license(s) of the initial developer.
  4.
    You may distribute machine-executable forms of the Software or machine-executable forms of modified versions of the Software, provided that you meet these restrictions:
    a. You must include this license document in the distribution.
    b. You must ensure that all recipients of the machine-executable forms are also able to receive the complete machine-readable source code to the distributed Software, including all modifications, without any charge beyond the costs of data transfer, and place prominent notices in the distribution explaining this.
    c. You must ensure that all modifications included in the machine-executable forms are available under the terms of this license.
  5.
    You may use the original or modified versions of the Software to compile, link and run application programs legally developed by you or by others.
  6.
    You may develop application programs, reusable components and other software items that link with the original or modified versions of the Software. These items, when distributed, are subject to the following requirements:
    a. You must ensure that all recipients of machine-executable forms of these items are also able to receive and use the complete machine-readable source code to the items without any charge beyond the costs of data transfer.
    b. You must explicitly license all recipients of your items to use and re-distribute original and modified versions of the items in both machine-executable and source code forms. The recipients must be able to do so without any charges whatsoever, and they must be able to re-distribute to anyone they choose.
    c. If the items are not available to the general public, and the initial developer of the Software requests a copy of the items, then you must supply one.

Limitations of Liability
In no event shall the initial developers or copyright holders be liable for any damages whatsoever, including - but not restricted to - lost revenue or profits or other direct, indirect, special, incidental or consequential damages, even if they have been advised of the possibility of such damages, except to the extent invariable law, if any, provides otherwise.

No Warranty
The Software and this license document are provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Choice of Law
This license is governed by the Laws of any member state of the European Union.
* * * * *  End Of File  * * * * */